from .device import Device
from .querier import Querier
from .linker import GenericLinker
from .dbmanager import GenericDBManager

class LightState:
    def __init__(self, dev):
        self._dev = dev
    
    def set_on(self, onLevel=1, port=None):
        self._dev.querier.query_std(0x11, max(0, min(255, int(255*onLevel))), port=port)

    def set_off(self, port=None):
        self._dev.querier.query_std(0x13, 0x00, port=port)

class Light(Device):
    def __init__(self, name, address, net=None, modem=None):
        super().__init__(name, address, net, modem)
        self.add_feature('querier', Querier(self))
        self.add_feature('db', GenericDBManager(self))
        self.add_feature('linker', GenericLinker(self))
        self.add_feature('state', LightState(self))



import threading
import traceback
import queue
import copy
import weakref

import time

from . import message
from .. import util as util

import logbook

logger = logbook.Logger(__name__)

class WriteRequest:
    def __init__(self, msg, priority=1, prewrite_quiet_time=0, postwrite_quiet_time=0,
            write_channel=None, first_reply_channel=None,
            ack_reply_channel=None, custom_channels=[], weakref_customs=True):
        self.msg = msg
        self.priority = priority

        self.prewrite_quiet_time = prewrite_quiet_time
        self.postwrite_quiet_time = postwrite_quiet_time

        self.write_channel = write_channel 

        self.first_reply_channel = first_reply_channel

        ack_msg_name = msg.type + 'Reply'

        # We must have this channel to know if we have received an ack!
        self.ack_reply_channel = ack_reply_channel if ack_reply_channel else \
                util.Channel(lambda x: x.type == ack_msg_name, 1)

        if not self.ack_reply_channel.has_filter or self.ack_reply_channel.has_activated:
            # No filter or if we are reusing a channel
            self.ack_reply_channel.reset_num_sent()
            self.ack_reply_channel.set_filter(lambda x: x.type == ack_msg_name)

        self.custom_channels = custom_channels # Will be added with write() call
        self.weakref_customs = weakref_customs

    def __gt__(self, other):
        return self.priority > other.priority
    def __lt__(self, other):
        return self.priority < other.priority
    def __ge__(self, other):
        return self.priority >= other.priority
    def __le__(self, other):
        return self.priority <= other.priority
    def __eq__(self, other):
        return self.priority == other.priority
    def __nq__(self, other):
        return self.priority != other.priority

# Takes a connection object that has read(), write(), flush(), and close() methods
class Port:
    def __init__(self, conn=None, definitions={}):
        self.defs = definitions

        # Connection and threads
        self._conn = None
        self._reader = None
        self._writer = None

        # A queue containing write requests
        self._write_queue = queue.PriorityQueue()

        # Setup the listeners
        self._read_listeners = []
        self._read_listeners_lock = threading.RLock()

        self._write_listeners = []
        self._write_listeners_lock = threading.RLock()

        # Setup the watchers, for optional use by the user
        # to print out to stdout the traffic through the port
        self._read_watcher = lambda x: print(str(x))
        self._write_watcher = lambda x: print(str(x))

        if conn:
            self.attach(conn)

    def __del__(self):
        self.detach()

    def attach(self, conn):
        if self._conn:
            self.detach()

        # Start threads
        self._conn = conn
        self._reader = threading.Thread(target=self._read_thread, daemon=True)
        self._writer = threading.Thread(target=self._write_thread, daemon=True)

        # Trigger for a thread to stop
        self._reader.stop = self.close
        self._writer.stop = self.close

        self._reader.start()
        self._writer.start()

    def detach(self):
        if self._reader:
            r = self._reader
            self._reader = None
            r.join()
        if self._writer:
            w = self._writer
            self._writer = None
            w.join()

        self._conn = None

    def close(self): # Detaches and closes the connection
        if self._conn:
            conn = self._conn
            self.detach()
            conn.close()

    def notify_on_read(self, handler):
        if not handler:
            return
        with self._read_listeners_lock:
            self._read_listeners.append(handler)

    def unregister_on_read(self, handler):
        if not handler:
            return
        with self._read_listeners_lock:
            if handler in self._read_listeners:
                self._read_listeners.remove(handler)

    def notify_on_write(self, handler):
        if not handler:
            return
        with self._write_listeners_lock:
            self._write_listeners.append(handler)

    def unregister_on_write(self, handler):
        if not handler:
            return
        with self._write_listeners_lock:
            if handler in self._write_listeners:
                self._write_listeners.remove(handler)

    # Utility debug functions.....

    def start_watching(self):
        self.notify_on_read(self._read_watcher)
        self.notify_on_write(self._write_watcher)
    
    def stop_watching(self):
        self.unregister_on_read(self._read_watcher)
        self.unregister_on_write(self._write_watcher)

    # Now the actual IO logic
    # added as weak references, so they won't stay
    # around if the channel goes out of scope by the original caller
    def write(self, msg, priority=1, prewrite_quiet_time=0, postwrite_quiet_time = 0,
            write_channel=None, first_reply_channel=None,
            ack_reply_channel=None, custom_channels=[], weakref_customs=True):

        self.write_request( WriteRequest(msg, priority,
            prewrite_quiet_time, postwrite_quiet_time,
            write_channel, first_reply_channel,
            ack_reply_channel, custom_channels, weakref_customs) )

    def write_request(self, request):
        self._write_queue.put(request)

    def _read_thread(self):
        decoder = message.MsgDecoder(self.defs)
        buf = bytes()
        while self._reader:
            try:
                buf = self._conn.read(1) # Read a byte
                # Feed into decoder
                msg = decoder.decode(buf)
                if not msg:
                    continue
            except TypeError as te: # Gets thrown on close() called during read() sometimes
                continue
            except Exception as e:
                print('Error reading!')
                print(traceback.format_exc())
                continue
            
            # Notify listeners
            with self._read_listeners_lock:
                listeners = list(self._read_listeners)
                for lr in listeners:
                    if isinstance(lr, weakref.ref):
                        l = lr()
                        if not l:
                            self._read_listeners.remove(lr)
                            continue
                    else:
                        l = lr
                    l(msg)

    def _write_thread(self):
        while self._writer:
            try:
                request = self._write_queue.get(timeout=0.1)
            except queue.Empty:
                continue

            # Wait the pre-write quiet time we want
            # during this time all is quiet
            if request.prewrite_quiet_time > 0:
                time.sleep(request.prewrite_quiet_time)

            msg = request.msg

            # Callback channels to notify on write/read:
            write_channel = request.write_channel
            first_reply_channel = request.first_reply_channel
            ack_reply_channel = request.ack_reply_channel

            if write_channel:
                write_channel.set_queuesize(1)
                write_channel.set_filter(None)

            if first_reply_channel:
                first_reply_channel.set_filter(None)

            # Our channels for writer control
            any_reply_channel = util.Channel() # For writer control
            any_reply_channel.set_filter(None) # Let everything in

            nack_reply_channel = util.Channel() # For writer control
            nack_reply_channel.set_filter(lambda msg: msg.type == 'PureNACK')

            # Get the listener lock so we don't miss anything
            # while we write
            with self._read_listeners_lock:
                # Add the channels
                self.notify_on_read(any_reply_channel)
                self.notify_on_read(nack_reply_channel)

                self.notify_on_read(first_reply_channel)
                self.notify_on_read(ack_reply_channel)

                self.notify_on_write(write_channel)

                # Will be removed by hand by the person who
                # queued the write if weakref_customs is False
                # otherwise will be automatically removed
                for channel in request.custom_channels:
                    if request.weakref_customs:
                        self.notify_on_read(weakref.ref(channel))
                    else:
                        self.notify_on_read(channel)

                for i in range(5): # Maximum of 5 resends...

                    # Now we do the actual writing
                    try: 
                        data = msg.bytes
                        self._conn.write(data)
                        self._conn.flush()
                    except Exception as e:
                        # TODO: Make logging
                        print('Error writing message!')
                        print(traceback.format_exc())
                        break # Move on to the next message

                    # Notify the listeners of the write
                    with self._write_listeners_lock:
                        listeners = list(self._write_listeners)
                        for lr in listeners:
                            if isinstance(lr, weakref.ref):
                                l = lr()
                                if not l:
                                    self._write_listeners.remove(lr)
                                    continue
                            else:
                                l = lr
                            l(msg)

                    # Remove the write channel the first time we write
                    if i == 0:
                        self.unregister_on_write(write_channel)

                    # Now we see what comes back by disabling the listener lock
                    self._read_listeners_lock.release()

                    resend = False
                    for mi in range(6): # Look at the next 6 messages or 0.6 second, max
                        if any_reply_channel.wait(0.1): # Wait 100ms for something to arrive
                            # Check if what came was an ack or nack
                            if ack_reply_channel.has_activated:
                                # Woo, we are done, no resend
                                break
                            elif nack_reply_channel.has_activated:
                                # Resend on a nack
                                any_reply_channel.clear()
                                nack_reply_channel.clear()
                                resend = True
                                break
                            else:
                                # Other message type...wait again
                                any_reply_channel.clear()

                    if not ack_reply_channel.has_activated and not resend:
                        # Resend due to no ack!
                        any_reply_channel.clear()
                        nack_reply_channel.clear()
                        resend = True

                    # Re-acquire so the with block doesn't get confused
                    self._read_listeners_lock.acquire() 

                    if not resend:
                        break

                # Remove the listeners
                self.unregister_on_read(first_reply_channel)
                self.unregister_on_read(ack_reply_channel)
                self.unregister_on_read(any_reply_channel)
                self.unregister_on_read(nack_reply_channel)

                # In case this hasn't happened
                self.unregister_on_write(write_channel)
            # Now outside of the lock,
            # post-write quiet time
            if request.postwrite_quiet_time > 0:
                time.sleep(request.postwrite_quiet_time)



import datetime

from .linkdb import LinkDB,LinkRecord
from . import linkdb
from ..util import Channel,InsteonError
from ..io.address import Address

import time

import logbook
logger = logbook.Logger(__name__)

class DBManager:
    def __init__(self, dev):
        self.cache = linkdb.LinkDB()
        self._dev = dev

    # Will update/return the target db,
    # create a new database if None
    def update_cache(self, targetdb=None, allow_linking=False, port=None):
        port = port if port else self._dev.port

        if not targetdb:
            targetdb = self.cache

        # _retrieve() returns a LinkDB() object
        records = self._retrieve(port)
        # check if based on the retrieved
        # records we would have the permissions
        # to get a proper db
        if not self._check_permissions(records):
            if not self._grant_permissions(allow_linking):
                logger.error('Not linked as a controller so cannot retrieve database. Use allow_linking to auto-link the modem as a controller of this device')
                raise InsteonError('Not linked as a controller so cannot access database')
            # Get a fresh copy now that we have permissions
            records = self._retrieve(port)
            if not self._check_permissions(records):
                raise InsteonError('Failed to get permission to modify the database')

        targetdb.update(records)

        return targetdb

    def flash_cache(self, srcdb=None, allow_linking=False, port=None):
        port = port if port else self._dev.port

        if not srcdb:
            srcdb = self.cache

        if not srcdb.valid:
            logger.error('The source database provided is not valid in that it has not have')
            logger.error('a timestamp and so is not considered "populated" to prevent accidentally')
            logger.error('wipping the database')
            raise InsteonError('Source database for flashing has not been set as populated')

        logger.trace('Getting latest linkdb')
        currentdb = linkdb.LinkDB()
        # Make a backup ....
        # Retrieve the current DB into a "currentdb" variable
        self.update_cache(targetdb=currentdb, allow_linking=allow_linking, port=port)

        # Save the currentdb
        backfile_name = '{}.linkdb.bk'.format(datetime.datetime.now().strftime('%b_%d_%Y_%H:%M:%S'))

        logger.warning('Modifying Link Database. This can go catastrophically wrong. A backup of the current database has been written to {}', backfile_name)
        currentdb.save(backfile_name)

        logger.trace('Writing linkdb changes')
        self._write(port, srcdb, currentdb)

    # Checks if the db has the permissions for
    # us to read
    def _check_permissions(self, currentdb):
        return True

    # Makes a database writable if the permission
    # check fails
    def _grant_permissions(self, allow_linking):
        return True

    # The actual implementation
    # that returns a database
    # retrieved
    def _retrieve(self, port):
        pass

    # The actual implementation that writes
    # a database given a current database and
    # a source database to write
    def _write(self, port, srcdb, currentdb):
        pass


#
# -----------------------------------------------------------------
# -------- Now specific database managers--------------------------
# -----------------------------------------------------------------
#

class GenericDBManager(DBManager):
    def __init__(self, dev):
        super().__init__(dev)

    def _check_permissions(self, currentdb):
        modem_addr = self._dev.modem.address

        # Check if the modem is in the currentdb
        # to see if we have the permissions to flash the database
        rec_filter = LinkRecord(address=modem_addr, flags_mask=(1 << 6))
        if rec_filter in currentdb:
            return True
        return False

    def _grant_permissions(self, allow_linking):
        if not allow_linking:
            return False
        # Link these things, check if there is a linker
        # feature for the modem and the device
        if not hasattr(self._dev, 'linker') or \
            not hasattr(self._dev.modem, 'linker'):
            raise InsteonError('Cannot link to device in order to read database since the device is missing the linker feature')

        logger.debug('Linking modem {} as controller to {}'.format(self._dev.modem.address,
                                                                    self._dev.address))

        self._dev.modem.linker.start_linking_controller()
        time.sleep(0.1) # Put a little sleep in there
        self._dev.linker.start_linking_responder()
        time.sleep(1) # Put a little sleep in there so the device has time to change its db
        return True

    def _retrieve(self, port):
        querier = self._dev.querier

        record_channel = Channel(lambda x: x.type == 'ExtendedMessageReceived' and \
                                            x['Cmd'] == 0x51 and \
                                            x['fromAddress'] == self._dev.address)
        response_channel = querier.query_ext(0x2f, 0x00, [], extra_channels=[record_channel])

        if not response_channel.recv(1):
            raise InsteonError('Did not get reply from device')

        db = linkdb.LinkDB()

        # Now keep receiving on the record channel until none are left
        while record_channel.wait(5):
            msg = record_channel.recv(5)
            offset = (msg['userData3'] & 0xFF) << 8 | (msg['userData4'] & 0xFF)
            flags = msg['userData6'] & 0xFF 
            group = msg['userData7'] & 0xFF
            address = Address(msg['userData8'], msg['userData9'], msg['userData10'])
            data = [msg['userData11'], msg['userData12'], msg['userData13']] 

            if not linkdb.LinkRecord(offset=offset) in db:
                db.add(linkdb.LinkRecord(offset, address, group, flags, data))

        return db

    def _write_entry(self, offset, record):
        record = record.copy()
        record.offset = None
        logger.debug('Writing record to {:04x}: {}'.format(offset, record))
        req_data = [0x00, 0x02]
        req_data.append((offset >> 8) & 0xFF)
        req_data.append(offset & 0xFF)
        req_data.append(8) # Set 8 bytes

        record_bytes = record.bytes
        # Make sure if nothing else that the high water
        # bit is set. If it isn't we could really brick the device
        record_bytes[0] = record_bytes[0] | 0x10
        req_data.extend(record_bytes)
        self._dev.querier.query_ext(0x2f, 0x00, req_data)
        time.sleep(1)

    def _null_entry(self, offset):
        logger.debug('Setting database end at {:04x}'.format(offset))
        req_data = [0x00, 0x02]
        req_data.append((offset >> 8) & 0xFF)
        req_data.append(offset & 0xFF)
        req_data.append(8) # Set 8 bytes
        req_data.extend(8*[0x00]) # null out the entry
        self._dev.querier.query_ext(0x2f, 0x00, req_data)
        time.sleep(1)

    def _write(self, port, srcdb, currentdb):
        # When nuking a database unlink the modem last
        modem_addr = self._dev.modem.address

        delete_records = []
        for record in currentdb:
            if not record.active:
                continue
            filter_rec = record.copy()
            filter_rec.offset = None # We don't care about offset
            if any(map(lambda x: x.offset < record.offset and filter_rec.matches(x), currentdb)):
                logger.debug('Found dup  record: {}'.format(record))
                delete_records.append(record)
            elif not filter_rec in srcdb:
                logger.debug('Found del  record: {}'.format(record))
                delete_records.append(record)


        free_records = list(delete_records)
        for record in currentdb:
            if not record.active:
                logger.debug('Found free record: {}'.format(record))
                free_records.append(record)
                if record.null:
                    # Append a bunch of dummy records after
                    for i in range(srcdb.size):
                        if record.offset - (i + 1) * 0x08 > 0x08:
                            free_records.append(LinkRecord(offset=record.offset - (i + 1)*0x08))

        logger.trace('Writing new records')

        # if the modem is being removed, add the modem at the first free offset temporarily
        modem_record = LinkRecord(address=modem_addr)
        modem_record.mask_link_type()
        modem_record.mask_active()
        modem_record.active = True
        modem_record.controller = False
        modem_record.high_water = False

        modem_offset = None
        if any(map(lambda x: modem_record.matches(x), delete_records)):
            if len(free_records) == 0:
                raise InsteonError('Out of database space!')
            modem_offset = free_records.pop(0).offset
            logger.trace('Writing modem record for editing into first free space')
            self._write_entry(modem_offset, modem_record)

        # Write any new entries into the free
        # spaces
        for record in srcdb:
            if not record.active:
                continue
            filter_rec = record.copy()
            filter_rec.offset = None # We don't care about offset
            if not filter_rec in currentdb:
                if len(free_records) == 0:
                    raise InsteonError('Out of database space!')
                free_offset = free_records.pop(0).offset
                self._write_entry(free_offset, record)

        logger.trace('Fetching fresh copy of database')
        currentdb = linkdb.LinkDB()
        self.update_cache(targetdb=currentdb, port=port)

        logger.trace('Disabling inactive records')

        # Disable anything that shouldn't be in the database
        for record in currentdb:
            if not record.active:
                continue
            if any(map(lambda x: record.matches(x), delete_records)):
                if modem_offset and record.offset == modem_offset:
                    continue # We'll handle this one at the end 
                new_record = record.copy()
                new_record.active = False
                self._write_entry(new_record.offset, new_record)

        logger.trace('Fetching fresh copy of database')
        currentdb = linkdb.LinkDB()
        self.update_cache(targetdb=currentdb, port=port)
        logger.trace('Setting end to null entry')

        # Chop off after the end
        self._null_entry(currentdb.end_offset)

        if modem_offset:
            logger.trace('Fetching fresh copy of database')
            currentdb = linkdb.LinkDB()
            self.update_cache(targetdb=currentdb, port=port)
            logger.trace('Disabling modem record')
            # If the modem is the last record, just wipe the database
            if modem_offset == currentdb.end_offset - 0x08:
                self._null_entry(modem_offset)
            else:
                new_record = currentdb.at(modem_offset).copy()
                new_record.active = False
                self._write_entry(new_record.offset, new_record)



class ModemDBManager(DBManager):
    def __init__(self, dev):
        super().__init__(dev)

    def _retrieve(self, port):
        reply_channel = Channel()
        done_channel = Channel(lambda x: (x.type == 'GetFirstALLLinkRecordReply' or \
                                              x.type == 'GetNextALLLinkRecordReply') and \
                                              x['ACK/NACK'] == 0x15)
        record_channel = Channel(lambda x: x.type == 'ALLLinkRecordResponse')


        # Now send the first message
        port.write(port.defs['GetFirstALLLinkRecord'].create(), ack_reply_channel=reply_channel,
                        custom_channels=[done_channel, record_channel])
        # Custom channels will be removed on garbage collect due to
        # weak references

        records = []

        while reply_channel.recv(5): # Wait at most 5 seconds for some reply
            if done_channel.has_activated: # If the reply says we are done, exit
                break
            # Wait another 2 seconds for the record
            msg = record_channel.recv(2)
            if not msg:
                raise InsteonError('No link data after ack for modem DB query')

            # Turn the msg into a record
            rec = linkdb.LinkRecord(None, msg['LinkAddr'], msg['ALLLinkGroup'],
                                    msg['RecordFlags'],
                                    [msg['LinkData1'], msg['LinkData2'], msg['LinkData3']])
            records.append(rec)

            # Request the next one
            port.write(port.defs['GetNextALLLinkRecord'].create(),
                        ack_reply_channel=reply_channel)
        else:
            raise InsteonError('Did not get reply for modem DB query')
        
        db = linkdb.LinkDB()
        db.update(records)

        return db

    def _write(self, port, srcdb, currentdb):
        for record in currentdb:
            filter_rec = record.copy()
            filter_rec.offset = None
            # If we do not find this record (regardless of offset) in the source
            # database, delete it!
            if not filter_rec in srcdb:
                logger.debug('Deleting record {}', record)

                msg = port.defs['ManageALLLinkRecord'].create()
                msg['controlCode'] = 0x80 # Delete by search
                msg['recordFlags'] = record.flags
                msg['ALLLinkGroup'] = record.group
                msg['linkAddress'] = record.address
                msg['linkData1'] = record.data[0]
                msg['linkData2'] = record.data[1]
                msg['linkData2'] = record.data[2]

                # Send the delete message and wait for a response
                ack_reply = Channel()
                port.write(msg, ack_reply_channel=ack_reply)
                reply_msg = ack_reply.recv(2)
                if reply_msg['ACK/NACK'] != 0x06:
                    raise InsteonError('The modem couldn\'t find the record we wanted to delete!')
                elif not reply_msg:
                    raise InsteonError('No reply to delete message')

        try:
            currentdb = linkdb.LinkDB()
            self.update_cache(targetdb=currentdb, port=port)
        except InsteonError as e:
            raise InsteonError('Unable to get database after removing records!') from e

        for record in srcdb:
            filter_rec = record.copy()
            filter_rec.offset = None
            if not filter_rec in currentdb:
                logger.debug('Adding record {}', record)
                msg = port.defs['ManageALLLinkRecord'].create()
                # Add resp. or controller
                msg['controlCode'] = 0x40 if (record.flags & (1 << 6)) else 0x41
                msg['recordFlags'] = record.flags
                msg['ALLLinkGroup'] = record.group
                msg['linkAddress'] = record.address
                msg['linkData1'] = record.data[0]
                msg['linkData2'] = record.data[1]
                msg['linkData3'] = record.data[2]

                ack_reply = Channel()
                port.write(msg, ack_reply_channel=ack_reply)
                if not ack_reply.wait(2):
                    raise InsteonError('No reply on record add!')



import threading
from contextlib import contextmanager

_bound_network = threading.local()

class Network:
    def __init__(self):
        self._dev_by_name = {}
        self._dev_by_addr = {}

    def register(self, device):
        if device.name:
            self._dev_by_name[device.name] = device
        self._dev_by_addr[device.address] = device

    def get_by_address(self, addr):
        if addr in self._dev_by_addr:
            return self._dev_by_addr[addr]
        else:
            return None

    def print(self):
        for name, dev in self._dev_by_name.items():
            print('{} {}'.format(dev.address.human, name))

    def bind(self):
        stack = getattr(_bound_network, 'stack', None)
        if not stack:
            stack = []
            _bound_network.stack = stack
        stack.append(self)

    def unbind(self):
        stack = getattr(_bound_network, 'stack', None)
        if stack:
            stack.remove(self)

    @contextmanager
    def use(self):
        self.bind()
        yield
        self.unbind()

    @staticmethod
    def bound():
        stack = getattr(_bound_network, 'stack', None)
        if stack:
            return stack[-1]
        else:
            return None




from . import address
from . import message
from .message import DataType
from .message import Direction

class FieldDef:
    def __init__(self, offset, length,
                 field_type,
                 name=None, default_value=None,
                 header_filter=None): # Header-filter is used for fields in the header
                                      # to see if a message matches a certain msg def
        self.offset = offset
        self.length = length
        self.type = field_type
        self.name = name
        self.default_value = default_value
        self.header_filter = header_filter

    def get(self, buf):
        o = self.offset
        if self.type == DataType.BYTE:
            return int.from_bytes(buf[o:o+1], byteorder='big')
        elif self.type == DataType.INT:
            return int.from_bytes(buf[o:o+4], byteorder='big')
        elif self.type == DataType.FLOAT:
            return struct.unpack('f', buf[o:o+4])[0]
        elif self.type == DataType.ADDRESS:
            return address.Address(int.from_bytes(buf[o:o+1], byteorder='big'),
                                   int.from_bytes(buf[o+1:o+2], byteorder='big'),
                                   int.from_bytes(buf[o+2:o+3], byteorder='big'))
        else:
            return None

    def set(self, buf, val):
        o = self.offset
        if self.type == DataType.BYTE:
            buf[o] = val.to_bytes(1, byteorder='big')[0]
        elif self.type == DataType.INT:
            buf[o:o+4] = val.to_bytes(4, byteorder='big')
        elif self.type == DataType.FLOAT:
            buf[o:o+4] = struct.pack('f', val)
        elif self.type == DataType.ADDRESS:
            buf[o:o+3] = val.bytes

    def format(self, msg=None):
        val = msg[self.name] if msg and self.name in msg else None

        if val is None:
            val = self.default_value

        if val is not None:
            valstr = str(val)
            if self.type == DataType.ADDRESS:
                valstr = val.human
            elif self.type == DataType.BYTE:
                valstr = '0x{:02x}'.format(val)
        else:
            valstr = '???'

        return self.name + ':' + valstr

class MsgDef:
    def __init__(self, name='', direction=Direction.TO_MODEM):
        self.name = name
        self.header_length = 0
        self.length = 0
        self.direction = direction
        self.fields_list = []
        self.fields_map = {}

    # Gets a field definition from
    # a message definition
    def __contains__(self, name):
        return name in self.fields_map

    def __getitem__(self, name):
        return self.fields_map[name]
    
    def append(self, field):
        self.length = self.length + field.length # Extend the length
        if field.name:
            self.fields_map[field.name] = field
        self.fields_list.append(field)

    # Check if the header matches
    def matches(self, buf):
        check_len = min(self.header_length, len(buf))
        for f in filter(lambda x: x.offset + x.type.value <= check_len and x.header_filter,
                                                                        self.fields_list):
            # Check the filter
            val = f.get(buf)
            if not (f.header_filter)(val):
                return False
        return True


    def create(self):
        return message.Msg(self)

    def deserialize(self, buf):
        m = message.Msg(self)
        for f in self.fields_list:
            if f.name is not None:
                val = f.get(buf)
                m[f.name] = val
        return m

    def serialize(self, msg):
        # Assume the 'type' field has already been set
        buf = bytearray(self.length)
        for f in self.fields_list:
            val = msg[f.name] if f.name is not None and \
                                    f.name in msg else \
                                    f.default_value
            if val is not None:
                f.set(buf, val)
        return bytes(buf)

    def format_msg(self, msg):
        sep = ('<','>') if self.direction == Direction.TO_MODEM else ('[',']')
        fields_str = '|'.join(map(lambda x: x.format(msg),
            filter(lambda x: x.name is not None, self.fields_list)))
        return sep[0] + self.name + sep[1] + ':' + fields_str

    def __str__(self):
        sep = ('<','>') if self.direction == Direction.TO_MODEM else ('[',']')
        fields_str = '|'.join(map(lambda x: x.format(),
            filter(lambda x: x.name is not None, self.fields_list)))
        return sep[0] + self.name + sep[1] + '(def):' + fields_str



from enum import Enum
import struct

class Direction(Enum):
    TO_MODEM = 'TO_MODEM'
    FROM_MODEM = 'FROM_MODEM'
    INVALID = 'INVALID'

class DataType(Enum):
    BYTE = 1
    INT = 4
    FLOAT = 4
    ADDRESS = 3
    INVALID = -1

class MsgType(Enum):
    BROADCAST = 0x80
    DIRECT = 0x00
    ACK_OF_DIRECT = 0x20
    NACK_OF_DIRECT = 0xa0
    ALL_LINK_BROADCAST = 0xc0
    ALL_LINK_CLEANUP = 0x40
    ALL_LINK_CLEANUP_ACK = 0x60
    ALL_LINK_CLEANUP_NACK = 0xe0

    def from_value(val):
        if not val:
            return None
        return MsgType(val & 0xe0)

    def from_msg(msg):
        if not 'messageFlags' in msg:
            return None

        return MsgType.from_value(msg['messageFlags'])

class AckType(Enum):
    ACK = 0x06
    NACK = 0x15

    def from_value(val):
        if not (val == 0x06 or val == 0x15):
            return None
        return AckType(val)

    def from_msg(msg):
        if not 'ACK/NACK' in msg:
            return None
        return AckType(msg['ACK/NACK'])

# Message behaves like a dictionary
# with a associated definition
class Msg:
    def __init__(self, msg_def, msg=None):
        self._def = msg_def
        self._msg = msg if msg else {} # The message dictionary

    def __contains__(self, name):
        return name in self._msg

    def __getitem__(self, name):
        return self._msg[name]

    def __setitem__(self, name, val):
        self._msg[name] = val

    def __str__(self):
        if self._def is None:
            return '<Unknown>'
        else:
            return self._def.format_msg(self)

    @property
    def type(self):
        return self._def.name

    @property
    def bytes(self):
        if self._def is None:
            return bytes()
        else:
            return self._def.serialize(self)

    def copy(self):
        return Msg(self._def,
                copy.deepcopy(self._msg))

class MsgDecoder:
    def __init__(self, defs = {}, direction=Direction.FROM_MODEM):
        self._buf = bytearray()
        self._direction = direction

        self._all_defs = defs.values()
        self._reset_filtered_defs()

    def _reset_filtered_defs(self):
        self._filtered_defs = list(filter(lambda d: d.direction == self._direction,
                                                        self._all_defs))

    def decode(self, buf):
        if len(self._filtered_defs) < 1:
            self._reset_filtered_defs()

        self._buf += buf

        while True:
            # A one-byte message isn't currently allowed!
            if len(self._buf) < 2:
                break
            
            # If we haven't yet determined the message, filter
            if len(self._filtered_defs) > 1: 
                self._filtered_defs = list(filter(lambda d: d.matches(self._buf), self._filtered_defs))
            # Now check to see if we are done or are totally lost
            if len(self._filtered_defs) == 1 and \
                    self._filtered_defs[0].length <= len(self._buf):
                # Done with message! Get the definition
                d = self._filtered_defs[0]
                # Pop off the required length and convert to immutable bytes
                msg_buf = bytes(self._buf[:d.length])
                self._buf = self._buf[d.length:]

                # Reset filtered defs
                self._reset_filtered_defs()

                return d.deserialize(msg_buf)
            elif len(self._filtered_defs) < 1:
                # Flush a single byte and reset the filter
                self._buf = self._buf[1:]

                self._reset_filtered_defs()
                continue
            # We can't determine anything yet
            break
        return None



from .device import Device
from .network import Network

from ..io.address import Address
from ..util import Channel

import threading
from contextlib import contextmanager

import logbook
logger = logbook.Logger(__name__)

_bound_modem = threading.local()

class Modem(Device):
    # Query for the address...
    def __init__(self, name, port, net=None):
        self._port = port

        addr = Address()
        # Query for the modem address
        addr_query = port.defs['GetIMInfo'].create()

        reply_channel = Channel()
        port.write(addr_query, ack_reply_channel=reply_channel)
        if reply_channel.wait(5): # Wait for a reply
            msg = reply_channel.recv()
            addr = msg['IMAddress']

        super().__init__(name, addr, net, self)

        # Add the features
        from .dbmanager import ModemDBManager
        self.add_feature('db', ModemDBManager(self))

        from .linker import ModemLinker
        self.add_feature('linker', ModemLinker(self))

    def bind(self):
        stack = getattr(_bound_modem, 'stack', None)
        if not stack:
            stack = []
            _bound_modem.stack = stack
        stack.append(self)

    def unbind(self):
        stack = getattr(_bound_modem, 'stack', None)
        if stack:
            stack.remove(self)

    @contextmanager
    def use(self):
        self.bind()
        yield
        self.unbind()

    @staticmethod
    def bound():
        stack = getattr(_bound_modem, 'stack', None)
        if stack:
            return stack[-1]
        else:
            return None



from xml.dom import minidom

from .message import Direction,DataType
from .message_def import FieldDef,MsgDef

def read_default_xml():
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore')
        import pkg_resources
        filepath = pkg_resources.resource_filename(__name__,'msg_definitions.xml')
        if not filepath:
            return None
        return read_xml(filepath)

def read_xml(filename):
    xmldoc = minidom.parse(filename)
    return process_xmltree(xmldoc)

# Returns a map of name to msgdef
def process_xmltree(xmldoc):
    results = {}

    elems = xmldoc.getElementsByTagName('msg')
    for m in elems:
        d = process_msgdef_elem(m)
        results[d.name] = d

    return results

# Returns a message
def process_msgdef_elem(elem):
    msg_name = elem.attributes['name'].value
    msg_length = int(elem.attributes['length'].value)
    msg_direction = Direction.FROM_MODEM \
                        if elem.attributes['direction'].value == 'FROM_MODEM' else \
                            Direction.TO_MODEM
    msg_def = MsgDef(msg_name, msg_direction)

    headers = elem.getElementsByTagName('header')
    if len(headers) > 1:
        raise ValueError('No header for message!')
    header = headers[0]
    for n in filter(lambda x: isinstance(x,minidom.Element),
                    header.childNodes):
        process_field_elem(n, msg_def)

    # Set the header length
    msg_def.header_length = msg_def.length

    for n in filter(
            lambda x: isinstance(x,minidom.Element) and x.tagName != 'header',
            elem.childNodes):
        process_field_elem(n, msg_def)

    if msg_length != msg_def.length:
        raise ValueError('Msg length and configured length not the same: {} vs {}'.format(msg_length,msg_def.length))
        
    return msg_def

def process_field_elem(field_elem, msgdef):
    name = None
    value = None
    filter_ = None

    if 'name' in field_elem.attributes:
        name = field_elem.attributes['name'].value

    if field_elem.firstChild and field_elem.firstChild.nodeValue:
        value = bytes.fromhex(field_elem.firstChild.nodeValue[2:])[0]

    if 'filter' in field_elem.attributes:
        filter_type = field_elem.attributes['filter'].value
        if filter_type == 'equals_default':
            filter_ = lambda x: x == value
        elif filter_type.startswith('bitset'):
            bit = int(filter_type.split(':')[1])
            filter_ = lambda x: x & (1 << bit) > 0
        elif filter_type.startswith('bitunset'):
            bit = int(filter_type.split(':')[1])
            filter_ = lambda x: x & (1 << bit) == 0

    typename = field_elem.tagName
    if typename == "byte":
        msgdef.append(FieldDef(msgdef.length, 1, DataType.BYTE,
                                name, value, filter_))
    elif typename == "int":
        msgdef.append(FieldDef(msgdef.length, 4, DataType.INT,
                                name, value, filter_))
    elif typename == "float":
        msgdef.append(FieldDef(msgdef.length, 4, DataType.FLOAT,
                                name, value, filter_))
    elif typename == "address":
        msgdef.append(FieldDef(msgdef.length, 3, DataType.ADDRESS,
                                name, value, filter_))
    else:
        raise ValueError('Unrecognized field type {}'.format(typename))



import insteon.io.message as msg
from ..io.address import Address

from . import network

import datetime
import json

import logbook
logger = logbook.Logger(__name__)

from warnings import warn

class LinkRecord:
    def __init__(self, offset=None, address=None, group=None, flags=None, data=None,
                        flags_mask=None):
        self.offset = offset
        self.address = address
        self.group = group
        self.flags = flags
        self.data = data
        self._flags_mask = flags_mask

    def mask_link_type(self):
        if not self._flags_mask:
            self._flags_mask = 0
        self._flags_mask |= (1 << 6)

    def mask_active(self):
        if not self._flags_mask:
            self._flags_mask = 0
        self._flags_mask |= (1 << 7)

    @property
    def null(self):
        return sum(self.bytes) == 0

    @property
    def active(self):
        if not self.flags:
            return False
        return self.flags & (1 << 7) > 0

    @active.setter
    def active(self, value):
        if not self.flags:
            self.flags = 0
        self.flags &= ~(1 << 7)
        if value:
            self.flags |= (1 << 7)

    @property
    def high_water(self):
        if not self.flags:
            return False
        return self.flags & (1 << 1) == 0

    @high_water.setter
    def high_water(self, value):
        if not self.flags:
            self.flags = 0
        self.flags &= ~(1 << 1)
        if not value:
            self.flags |= (1 << 1)

    @property
    def controller(self):
        if not self.flags:
            return False
        return self.flags & (1 << 6) > 0

    @controller.setter
    def controller(self, value):
        if not self.flags:
            self.flags = 0
        self.flags &= ~(1 << 6)
        if value:
            self.flags |= (1 << 6)

    @property
    def responder(self):
        return not self.controller

    @responder.setter
    def responder(self, value):
        self.controller = not value

    @property
    def bytes(self):
        return [(self.flags if self.flags else 0), \
                (self.group if self.group else 0)] + \
                self.address.array + \
                (self.data if self.data else [0, 0, 0])

    @property
    def packed(self):
        if self.offset:
            return {'offset': self.offset, 'address': self.address.packed,
                    'group': self.group, 'flags': self.flags, 'data': self.data}
        else:
            return {'address': self.address.packed,
                    'group': self.group, 'flags': self.flags, 'data': self.data}

    @staticmethod
    def unpack(packed):
        if 'offset' in packed:
            return LinkRecord(packed['offset'], Address.unpack(packed['address']),
                              packed['group'], packed['flags'], packed['data'])
        else:
            return LinkRecord(None, Address.unpack(packed['address']),
                              packed['group'], packed['flags'], packed['data'])

    def copy(self):
        return LinkRecord(self.offset, self.address, self.group,
                            self.flags, self.data)

    def matches(self, other):
        if self.offset is not None \
                and other.offset is not None \
                and self.offset != other.offset:
            return False
        if self.address is not None \
                and other.address is not None \
                and self.address != other.address:
            return False
        if self.group is not None \
                and other.group is not None \
                and self.group != other.group:
            return False
        if self.flags is not None \
                and other.flags is not None:
            filter_flags = self._flags_mask if self._flags_mask else other._flags_mask
            if filter_flags and self.flags & filter_flags != other.flags & filter_flags:
                return False
            if not filter_flags and self.flags != other.flags:
                return False
        if self.data is not None \
                and other.data is not None \
                and self.data != other.data:
            return False
        return True

    def __str__(self):
        flags = self.flags if self.flags else 0
        valid = (flags & (1 << 7))
        ltype = 'CTRL' if (flags & (1 << 6)) else 'RESP'
        ctrl = ' ' + ltype + ' ' if valid else '(' + ltype + ')'
        data_str = ' '.join([format(x & 0xFF, '02x') \
                             for x in (self.data if self.data else [0, 0, 0])])
        dev = self.address.human if self.address else 'xx.xx.xx'
        addr = self.address.human if self.address else 'xx.xx.xx'
        offset = self.offset if self.offset else 0xffff
        group = self.group if self.group else 0

        if network.Network.bound():
            device = network.Network.bound().get_by_address(self.address)
            if device:
                dev = device.name

        if self.offset:
            return '{:04x} {:30s} {:8s} {} {:08b} group: {:02x} data: {}'.format(
                    offset, dev, addr, ctrl, flags, group, data_str)
        else:
            return '{:30s} {:8s} {} {:08b} group: {:02x} data: {}'.format(
                    dev, addr, ctrl, flags, group, data_str)


class LinkDB:
    def __init__(self, records=None, timestamp=None):
        self.records = records if records else []
        self.timestamp = timestamp

    def __iter__(self):
        for r in self.records:
            yield r

    def __contains__(self, record):
        for r in self.records:
            if record.matches(r):
                return True
        return False

    @property
    def empty(self):
        return not self.records

    @property
    def valid(self):
        return self.timestamp is not None

    @property
    def size(self):
        return len(self.records)

    @property
    def start_offset(self):
        return 0x0fff

    @property
    def end_offset(self):
        last_off = 0x0fff
        for r in self.records:
            if not r.active:
                continue
            if r.offset and r.offset - 0x08 < last_off:
                last_off = r.offset - 0x08
        return last_off

    # Sets the timestamp of the device (if ts is None, sets it to the current time)
    def set_timestamp(self, ts=None):
        self.timestamp = ts if ts else datetime.datetime.now()

    def set_invalid(self):
        self.timestamp = None

    # Editing commands

    def at(self, offset):
        for r in self.records:
            if r.offset == offset:
                return r
        return None

    def remove(self, rec):
        self.records.remove(rec)

    def add(self, rec):
        if not rec in self.records:
            self.records.append(rec)

    def clear(self):
        self.records.clear()

    # For removing a particular device
    def remove_matching(self, filter_rec):
        for r in list(self.records):
            if filter_rec.matches(r):
                self.remove(r)

    def remove_device(self, dev):
        addr = dev.address
        self.remove_matching(LinkRecord(address=addr))

    def remove_address(self, address):
        self.remove_matching(LinkRecord(address=address))

    def remove_offset(self, offset):
        self.remove_matching(LinkRecord(offset=offset))

    # For adding new records
    def add_address(self, address, controller=False, group=0x01, data=[0, 0, 0]):
        rec = LinkRecord(address=address, group=0x01, data=data)
        rec.active = True
        rec.high_water = False
        rec.controller = controller
        self.add(rec)

    def add_device(self, device, controller=False, group=0x01, data=[0, 0, 0]):
        self.add_address(device.address, controller, group, data)

    def copy(self):
        return LinkDB([x.copy() for x in self.records], self.timestamp)

    # For filtering by a record
    # will return a new database
    def filter(self, filter_rec):
        rec = []
        for r in self.records:
            if filter_rec.matches(r):
                rec.append(r)
        db = LinkDB(rec, self.timestamp)
        return db

    # For serialization/unserialization
    @property
    def packed(self):
        packed = {}
        packed['timestamp'] = self.timestamp.strftime('%b %d %Y %H:%M:%S')
        records = []
        for r in self.records:
            records.append(r.packed)
        packed['records'] = records
        return packed

    @staticmethod
    def unpack(packed):
        timestamp = None
        if 'timestamp' in packed:
            timestamp = datetime.datetime.strptime(packed['timestamp'], '%b %d %Y %H:%M:%S')
        records = []
        if 'records' in packed:
            for r in packed['records']:
                records.append(LinkRecord.unpack(r))
        return LinkDB(records, timestamp)

    def load(self, filename):
        with open(filename, 'r') as i:
            packed = json.load(i)
            self.update(LinkDB.unpack(packed))

    def save(self, filename):
        with open(filename, 'w') as out:
            json.dump(self.packed, out)

    # Adds a bunch of records and sets the timestamp (if records has a timestamp property, it uses
    # that, otherwise it just uses the current time)
    def update(self, records):
        self.clear()
        for r in records:
            self.add(r)

        # If we are looking at another database
        # it will have a timestamp on it
        if hasattr(records, 'timestamp'):
            self.set_timestamp(records.timestamp)
        else:
            self.set_timestamp()

    def print(self, formatter=None):
        if not self.valid:
            logger.warning('LinkDB cache not valid!')
            return

        print(self.timestamp.strftime('Retrieved: %b %d %Y %H:%M:%S'))
        for rec in self.records:
            print(rec)



from ..util import Channel,InsteonError, calc_simple_crc, calc_long_crc
from ..io.message import MsgType

class Querier:
    def __init__(self, dev):
        self._dev = dev

    def send_std(self, cmd1, cmd2, flag=MsgType.DIRECT,
                        wait_response=False, extra_channels=[], port=None):
        port = port if port else self._dev.port

        msg = port.defs['SendStandardMessage'].create()
        msg['toAddress'] = self._dev.address
        msg['messageFlags'] = flag.value | 0xf
        msg['command1'] = cmd1
        msg['command2'] = cmd2

        direct_ack_channel = Channel(lambda x: x.type == 'StandardMessageReceived' and
                                            (MsgType.from_value(x['messageFlags']) == MsgType.ACK_OF_DIRECT or
                                             MsgType.from_value(x['messageFlags']) == MsgType.NACK_OF_DIRECT) and
                                            x['fromAddress'] == self._dev.address)
        ack_reply = Channel()

        extras = [direct_ack_channel]
        extras.extend(extra_channels)

        port.write(msg, ack_reply_channel=ack_reply, custom_channels=extras)

        if not ack_reply.wait(3):
            raise InsteonError('No IM reply to send command!')

        if wait_response:
            if not direct_ack_channel.wait(4):
                raise InsteonError('No response to standard query received')

        return direct_ack_channel

    def query_std(self, cmd1, cmd2, flag=MsgType.DIRECT,
                    wait_response=True, extra_channels=[], port=None):
        return self.send_std(cmd1, cmd2, flag, wait_response, extra_channels, port)

    def send_ext(self, cmd1, cmd2, data, flag=MsgType.DIRECT, large_checksum=False,
                    wait_response=False, extra_channels=[], port=None):
        port = port if port else self._dev.port

        msg = port.defs['SendExtendedMessage'].create()
        msg['toAddress'] = self._dev.address
        msg['messageFlags'] = flag.value | (1 << 4) | 0xf
        msg['command1'] = cmd1
        msg['command2'] = cmd2

        if large_checksum and len(data) > 12 or len(data) > 13:
            raise InsteonError('Cannot send more than 12 or 13 bytes in an ext message')

        for i,x in enumerate(data):
            msg['userData{}'.format(i + 1)] = x

        checksum_data = [cmd1, cmd2]
        checksum_data.extend(data)

        if large_checksum:
            crc = calc_long_crc(checksum_data)
            msg['userData13'] = crc & 0xFF
            msg['userData14'] = (crc >> 8) & 0xFF
        else:
            msg['userData14'] = calc_simple_crc(checksum_data)

        direct_ack_channel = Channel(lambda x: (x.type == 'ExtendedMessageReceived' or 
                                                x.type == 'StandardMessageReceived') and
                                            (MsgType.from_value(x['messageFlags']) == MsgType.ACK_OF_DIRECT or
                                             MsgType.from_value(x['messageFlags']) == MsgType.NACK_OF_DIRECT))
        ack_reply = Channel()

        extras = [direct_ack_channel]
        extras.extend(extra_channels)

        port.write(msg, ack_reply_channel=ack_reply, custom_channels=extras)

        if not ack_reply.wait(3):
            raise InsteonError('No IM reply to send command!')

        if wait_response:
            if not direct_ack_channel.wait(4):
                raise InsteonError('No response to extended query received')

        return direct_ack_channel

    def query_ext(self, cmd1, cmd2, data, flag=MsgType.DIRECT, large_checksum=False,
                    wait_response=True, extra_channels=[], port=None):
        return self.send_ext(cmd1, cmd2, data, flag, large_checksum, wait_response, extra_channels, port)



class SerialConn:
    FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS = 5, 6, 7, 8
    STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO = 1, 1.5, 2
    PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 'N', 'E', 'O', 'M', 'S'

    def __init__(self, port, baudrate=19200, bytesize=EIGHTBITS, parity=PARITY_NONE,
                    stopbits=STOPBITS_ONE, timeout=0.1,
                    xonxoff=False, rtscts=False, write_timeout=None, dsrdtr=False,
                    inter_byte_timeout=None):
        # Find the right serial implementation to use
        import serial
        self._port = serial.Serial(port=port, baudrate=baudrate,
                                    bytesize=bytesize, parity=parity,
                                    stopbits=stopbits, timeout=timeout,
                                    xonxoff=xonxoff, rtscts=rtscts,
                                    write_timeout=write_timeout, dsrdtr=dsrdtr,
                                    inter_byte_timeout=inter_byte_timeout)

    def __del__(self):
        if self.is_open:
            self.close()
    
    @property
    def is_open(self):
        return self._port.is_open

    def close(self):
        self._port.close()

    def read(self, size=1):
        return self._port.read(size)

    def write(self, data):
        return self._port.write(data)

    def flush(self):
        self._port.flush()



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='insteon',  # Required
    version='0.0.1',  # Required
    description='A utility for interacting with insteon modems and hubs',
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/pfrommerd/python-insteon',
    author='Daniel Pfrommer',
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'
    ],

    keywords='insteon',  # Optional
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'deploy', 'target']),  # Required

    install_requires=['pyserial','logbook','setuptools'],  # Optional
    extras_require = {  # Optional
        'dev': [],
        'test': []
    },

    package_data = {  # For in-package data files
        'insteon': ['io/msg_definitions.xml'],
    },

    data_files=[],

    project_urls={  
        'Source': 'https://github.com/pfrommerd/python-insteon/',
    },
)



from http.server import BaseHTTPRequestHandler, HTTPServer

import insteon.io.serial as serial

import binascii
import threading

def make_handler(srv):
    class HubHandler(BaseHTTPRequestHandler):
        def do_HEAD(self):
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()

        def do_GET(self):
            # Parse the path to either
            if self.path == '/buffstatus.xml':
                print('Getting buffer status')
                self.handle_buffer_request()
            elif self.path == '/1?XB=M=1':
                print('Clearing buffer')
                self.handle_buffer_clear()
            elif self.path.startswith('/3?') and self.path.endswith('=I=3'):
                data = self.path[3:len(self.path) - 4]
                print('Writing: {}'.format(data))
                self.handle_write_request(data)
            else:
                print('Unhandled request: {}'.format(self.path))
                self.send_response(404)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.send_text('Not found!')

        def log_message(self, format, *args):
            pass

        def handle_buffer_request(self):
            self.send_response(200)
            self.send_header('Content-type', 'text/xml')
            self.end_headers()
            data = srv.buffer_status
            self.send_text('<response><BS>' + data + '</BS></response>')

        def handle_buffer_clear(self):
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            srv.clear()
            self.send_text('buffer cleared!')

        def handle_write_request(self, hex_data):
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            try:
                data = binascii.unhexlify(hex_data)
                self.send_text('writing: ' + data.hex())
                srv.write(data)
            except:
                print('Error writing: {}'.format(hex_data))
                self.send_text('error!')
                return

        def send_text(self, text):
            self.wfile.write(bytes(text, 'UTF-8'))
    return HubHandler

class HubServer:
    def __init__(self, io_conn, port, username, password, bufferlen):
        self._io_conn = io_conn
        self._port = port
        self._handler = make_handler(self)

        self._buffer_lock = threading.Lock()
        self._buffer = bytearray(bufferlen)
        self._buffer_pos = 0

    @property
    def buffer(self):
        with self._buffer_lock:
            return bytes(self._buffer)

    @property
    def buffer_status(self):
        with self._buffer_lock:
            return (self._buffer.hex() + '{:02x}'.format(2*self._buffer_pos)).upper()


    def write(self, data):
        with self._buffer_lock:
            self._io_conn.write(data)

    def read(self, data):
        with self._buffer_lock:
            for x in data:
                self._buffer[self._buffer_pos] = x
                self._buffer_pos = (self._buffer_pos + 1) % len(self._buffer)

    def clear(self):
        with self._buffer_lock:
            self._buffer = bytearray(len(self._buffer))
            self._buffer_pos = 0

    def run(self):
        # start the read thread
        reader = threading.Thread(target=self._read_thread)
        reader.start()

        self._run_http_server()

        self._io_conn.close()
        reader.join()

    def _run_http_server(self):
        httpd = HTTPServer(("", self._port), self._handler)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print('Received interrupt, shutting down server')
        httpd.server_close()

    def _read_thread(self):
        while self._io_conn.is_open:
            buf = self._io_conn.read(1)
            if len(buf) > 0:
                self.read(buf)

def run():
    port = 25105 
    username = 'hub'
    password = 'hub'
    print('Server started on port {}'.format(port))
    srv = HubServer(serial.SerialConn('/dev/ttyUSB0'), port, username, password, 40)
    srv.run()

if __name__ == '__main__':
    run()




class Address:
    def __init__(self, hi=0, mid=0, low=0):
        self._hi = hi;
        self._mid = mid;
        self._low = low;

    @property
    def bytes(self):
        return bytes([self._hi, self._mid, self._low])

    @property
    def array(self):
        return [self._hi, self._mid, self._low]

    @property
    def human(self):
        return '{:02x}.{:02x}.{:02x}'.format(self._hi, self._mid, self._low).upper()

    @property
    def packed(self):
        return [self._hi, self._mid, self._low]

    @staticmethod
    def unpack(packed):
        return Address(packed[0], packed[1], packed[2])

    def __str__(self):
        return self.human

    def __eq__(self, other):
        return self._hi == other._hi and self._mid == other._mid and self._low == other._low

    def __nq__(self, other):
        return self._hi != other._hi or self._mid != other._mid or self._low != other._low

    def __hash__(self):
        return 13 * self._hi + 57 * self._mid + 29*self._low



class Device:
    def __init__(self, name, addr, net=None, modem_=None):
        from . import network
        from . import modem

        self._name = name
        self._addr = addr
        self._features = {}
        self._network = net if net else network.Network.bound()
        self._modem = modem_ if modem_ else modem.Modem.bound()

        if self._network:
            self._network.register(self)

    @property
    def name(self):
        return self._name

    @property
    def address(self):
        return self._addr

    @property
    def features(self):
        return self._features

    @property
    def network(self):
        return self._network

    @property
    def modem(self):
        return self._modem

    @property
    def port(self):
        if self._modem:
            return self._modem._port
        else:
            return None

    def add_feature(self, name, feature):
        setattr(self, name, feature)
        self.features[name] = feature



from time import time as _time
import queue
import threading
import inspect

# CRC stuff

def calc_simple_crc(data):
    return (~(sum(data)) + 1) & 0xFF

def calc_long_crc(data):
    crc = int(0);
    for loop in xrange(0, len(bytes)):
            b = bytes[loop] & 0xFF
            for bit in xrange(0, 8):
                    fb = b & 0x01
                    fb = fb ^ 0x01 if (crc & 0x8000) else fb
                    fb = fb ^ 0x01 if (crc & 0x4000) else fb
                    fb = fb ^ 0x01 if (crc & 0x1000) else fb
                    fb = fb ^ 0x01 if (crc & 0x0008) else fb
                    crc = ((crc << 1) | fb) & 0xFFFF;
                    b = b >> 1
    return crc

class Channel:
    def __init__(self, filter=None, maxqueue=0):
        self._queue = queue.Queue(maxqueue)

        self._counter = 0 # Number of messages queued
        self._counter_lock = threading.Lock()

        self._filter = filter

    @property
    def has_filter(self):
        return self._filter is not None

    @property
    def active(self):
        return not self._queue.empty()

    @property
    def has_activated(self):
        return self.num_sent > 0

    @property
    def num_sent(self):
        with self._counter_lock:
            return self._counter

    def reset_num_sent(self):
        with self._counter_lock:
            self._counter = 0

    def clear(self):
        with self._queue.mutex:
            self._queue.queue.clear()

    def set_queuesize(self, newsize):
        # oops, just for get about the current
        # elements for now
        self._queue = queue.Queue(newsize)

    def set_filter(self, filter):
        self._filter = filter

    # Will stack another filter ontop of the current one
    def chain_filter(self, filter):
        if self._filter:
            f = self._filter
            def chain(*args):
                return f(*args) and filter(*args)
            self._filter = chain
        else:
            self._filter = filter

    # Waits until there is at least one element in the queue
    # but doesn't take it
    def wait(self, timeout=None):
        with self._queue.not_empty:
            if timeout is None:
                if not self._queue._qsize():
                    self._queue.not_empty.wait()
                return True
            elif timeout > 0:
                endtime = _time() + timeout
                while not self._queue._qsize():
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        return False
                    self._queue.not_empty.wait(remaining)
                return True
            elif timeout <= 0:
                # Just return whether there
                # is anything in the queue
                return self._queue._qsize()

    def recv(self, timeout=None):
        try:
            if timeout is None:
                return self._queue.get()
            elif timeout > 0:
                return self._queue.get(True, timeout)
            elif timeout <= 0:
                return self._queue.get(False)
        except queue.Empty:
            return None

    def send(self, *args):
        # Push args tuple onto queue
        # if only a single arg, push the first arg
        arg = args
        if len(arg) == 1:
            arg = args[0]
        elif len(arg) == 0:
            arg = None

        try:
            if not self._filter or self._filter(*args):
                with self._counter_lock:
                    self._counter = self._counter + 1
                    self._queue.put(arg)
        except queue.Full:
            return # Drop...

    def __call__(self, *args):
        self.send(*args)

# A custom insteon error type
# that the terminal knows can just be printed out
# without a stack trace
# this should be used for when messages are dropped
# (i.e any unexpected behavior on the insteon network
# or with the modem)
class InsteonError(Exception):
    def __init__(self, *args):
        super().__init__(*args)
        self.quiet = True



from ..util import Channel,InsteonError

import logbook
logger = logbook.Logger(__name__)

class Linker:
    def __init__(self, dev):
        self._dev = dev

    def start_linking_responder(self, group=0x01, port=None):
        pass

    def start_linking_controller(self, group=0x01, port=None):
        pass

    def stop_linking(self, port=None):
        pass

class ModemLinker(Linker):
    def __init__(self, modem):
        super().__init__(modem)

    def start_linking_responder(self, group=0x01, port=None):
        port = port if port else self._dev.port
        if not port:
            raise InsteonError('No port specified')

        msg = port.defs['StartALLLinking'].create()
        msg['LinkCode'] = 0x00
        msg['ALLLinkGroup'] = group

        ack_reply = Channel()
        port.write(msg, ack_reply_channel=ack_reply)

        if not ack_reply.wait(1):
            raise InsteonError('Received no reply')

    def start_linking_controller(self, group=0x01, port=None):
        port = port if port else self._dev.port
        if not port:
            raise InsteonError('No port specified')

        msg = port.defs['StartALLLinking'].create()
        msg['LinkCode'] = 0x01
        msg['ALLLinkGroup'] = group

        ack_reply = Channel()
        port.write(msg, ack_reply_channel=ack_reply)

        if not ack_reply.wait(1):
            raise InsteonError('Received no reply')

    def stop_linking(self, port=None):
        port = port if port else self._dev.port
        if not port:
            raise InsteonError('No port specified')

        msg = port.defs['CancelALLLinking'].create()
        ack_reply = Channel()
        port.write(msg, ack_reply_channel=ack_reply)

        if not ack_reply.wait(1):
            raise InsteonError('Received no reply')

class GenericLinker(Linker):
    def __init__(self, dev):
        super().__init__(dev)

    def start_linking_controller(self, group=0x01, port=None):
        port = port if port else self._dev.port
        if not port:
            raise InsteonError('No port specified')
        self._dev.querier.query_ext(0x09, 0x01, [])

    def start_linking_responder(self, group=0x01, port=None):
        self.start_linking_controller(group, port)

    def stop_linking(self, port=None):
        raise InsteonError('Not implemented')
